from typing import NewType, assert_type
from uuid import UUID

from lastuuid.factories import (
    LastUUIDFactory,
    NewTypeFactory,
)


def test_newtype_factory():
    UserId = NewType("UserId", UUID)
    newtype = NewTypeFactory[UserId](UserId)
    val = newtype()
    assert_type(val, UserId)


def test_last_uuid_factory():
    ClientId = NewType("ClientId", UUID)

    client_id_factory = LastUUIDFactory[ClientId](ClientId, cache_size=2)
    myid = client_id_factory()
    assert myid == client_id_factory.last
    myid2 = client_id_factory()
    assert myid2 == client_id_factory.last
    assert [myid2, myid] == client_id_factory.lasts
    myid3 = client_id_factory()
    assert myid3 == client_id_factory.last
    assert [myid3, myid2] == client_id_factory.lasts



from uuid import UUID

from lastuuid.dummies import uuid7gen, uuidgen


def test_default():
    assert uuidgen() == UUID(int=1)  # XXX this tests works because it is the first
    assert uuidgen.last == UUID(int=1)


def test_predictable():
    assert uuidgen(1) == UUID("00000001-0000-0000-0000-000000000000")
    assert uuidgen(1, 2, 3, 4, 5) == UUID("00000001-0002-0003-0004-000000000005")


def test_predictable_uuid7():
    myid = uuid7gen()
    assert myid == uuid7gen.last



from datetime import UTC, date, datetime
from uuid import UUID

from lastuuid.utils import (
    uuid7_bounds_from_date,
    uuid7_bounds_from_datetime,
)


def test_uuid7_range_from_datetime():
    dt = datetime(2025, 2, 14, 0, 0, 0, tzinfo=UTC)
    left, right = uuid7_bounds_from_datetime(dt)
    assert left == UUID("019501c1-4c00-7000-b000-000000000000")
    assert right == UUID("019501c1-4c01-7000-b000-000000000000")


def test_uuid7_range_from_datetime_2():
    dt1 = datetime(2025, 2, 14, 0, 0, 0, tzinfo=UTC)
    dt2 = datetime(2025, 2, 15, 0, 0, 0, tzinfo=UTC)
    left, right = uuid7_bounds_from_datetime(dt1, dt2)
    assert left == UUID("019501c1-4c00-7000-b000-000000000000")
    assert right == UUID("019506e7-a800-7000-b000-000000000000")


def test_uuid7_range_from_date():
    dt1 = date(2025, 2, 14)
    left, right = uuid7_bounds_from_date(dt1)
    assert left == UUID("019501c1-4c00-7000-b000-000000000000")
    assert right == UUID("019506e7-a800-7000-b000-000000000000")



#!/usr/bin/env python3
import datetime
from importlib.metadata import version

header = (
    f"## {version('lastuuid')}  - " f" {datetime.datetime.now().date().isoformat()}"
)
with open("CHANGELOG.md.new", "w") as changelog:
    changelog.write(header)
    changelog.write("\n")
    changelog.write("\n")
    changelog.write("* please write here \n\n")



"""
A dummy uuid usefull for unit testing purpose.

UUID generated here are full of 0, they does not respect any kind of UUID version,
they remove a bit of cognitive load while testing.
"""

from typing import Iterator
from uuid import UUID

from lastuuid.factories import LastUUIDFactory

__all__ = ["uuidgen", "uuid7gen"]


def gen_id() -> Iterator[int]:
    num = 0
    while True:
        num += 1
        yield num


next_id = gen_id()


def _uuidgen(i: int = 0, j: int = 0, k: int = 0, x: int = 0, y: int = 0) -> UUID:
    if i == 0 and y == 0:
        y = next(next_id)
    return UUID(f"{i:0>8}-{j:0>4}-{k:0>4}-{x:0>4}-{y:0>12}")


uuidgen = LastUUIDFactory(None, _uuidgen)
"""
A UUID generator that makes UUIDs more readable for humans.

# Generate autoincrement UUID that you need to predict

Sometime you prepare fixtures with well known UUID and you want to repeat them,

uuidgen(1) is more readable that UUID('00000001-0000-0000-0000-000000000000'),
this is why this function is made for.
Every section of the uuid can be filled out using `i`, `j`, `k`, `x`, `y` but,
I personnaly never use more than `i` and `j`.

```python
>>> from lastuuid.dummies import uuidgen
>>> uuidgen(1)
UUID('00000001-0000-0000-0000-000000000000')
>>> uuidgen(1, 2)
UUID('00000001-0002-0000-0000-000000000000')
>>> uuidgen(1, 2, 3, 4, 5)
UUID('00000001-0002-0003-0004-000000000005')
```

```{tip}
if you don't want a dependency for that, the standard library let you write
UUID(int=1) which produce UUID('00000000-0000-0000-0000-000000000001').
```

# Generate autoincrement UUID that you don't need to predict

Without any parameter, it will generate UUID where the last bits are incremented.
It also keep the 10 lasts created values.

```python
>>> from lastuuid.dummies import uuidgen
>>> uuidgen()
UUID('00000000-0000-0000-0000-000000000001')
>>> uuidgen.last
UUID('00000000-0000-0000-0000-000000000001')
>>> uuidgen()
UUID('00000000-0000-0000-0000-000000000002')
>>> uuidgen.last
UUID('00000000-0000-0000-0000-000000000002')
>>> uuidgen.lasts[1]
UUID('00000000-0000-0000-0000-000000000001')
```

"""

uuid7gen = LastUUIDFactory(None)
"""
Generate uuid7 and store the last generated values to get them
using last and lasts property.

```python
>>> from lastuuid.dummies import uuid7gen
>>> uuid7gen()
UUID('019b5cb2-2fea-7572-8fc6-84fc550278e8')
>>> uuid7gen.last
UUID('019b5cb2-2fea-7572-8fc6-84fc550278e8')
>>> uuid7gen.lasts
[UUID('019b5cb2-2fea-7572-8fc6-84fc550278e8')]
>>> uuid7gen()
UUID('019b5cb2-82e5-709d-a24d-7dba31eb3e82')
>>> uuid7gen.last
UUID('019b5cb2-82e5-709d-a24d-7dba31eb3e82')
>>> uuid7gen.last[0]
UUID('019b5cb2-82e5-709d-a24d-7dba31eb3e82')
>>> uuid7gen.lasts[1]
UUID('019b5cb2-2fea-7572-8fc6-84fc550278e8')
```

See also {class}`lastuuid.factories.LastUUIDFactory`.
"""



"""Pyo3 binding interface definition."""

from uuid import UUID
from datetime import datetime, timezone


def uuid7() -> UUID:
    """
    Generate an uuid using uuidv7 format, the best format that feet in a BTree.
    """


def uuid7_to_datetime(uuid7: UUID, tz=timezone.utc) -> datetime:
    """Extract the datetime part of an uuid 7."""



from uuid import UUID
from datetime import datetime, timedelta, timezone

import pytest
from pydantic import BaseModel, Field

from lastuuid import uuid7, uuid7_to_datetime


dummy_timezone = timezone(timedelta(hours=2))


class Dummy(BaseModel):
    id: UUID = Field(default_factory=uuid7)


def test_dummy():
    dummy = Dummy()
    dummy2 = Dummy()
    assert dummy.id.bytes < dummy2.id.bytes

    assert dummy.model_dump(mode="json") == {"id": str(dummy.id)}


def test_uuid7_to_datetime_default():
    assert uuid7_to_datetime(UUID("019500d0-468f-7eaa-ab66-f90e998cd72c")) == datetime(
        2025, 2, 13, 19, 36, 44, 431000, tzinfo=timezone.utc
    )


def test_uuid7_to_datetime_naive_custom_timezone():
    assert uuid7_to_datetime(
        UUID("019500d0-468f-7eaa-ab66-f90e998cd72c"), tz=dummy_timezone
    ) == datetime(2025, 2, 13, 21, 36, 44, 431000, tzinfo=dummy_timezone)


def test_uuid7_to_datetime_uuid4():
    with pytest.raises(ValueError) as ctx:
        uuid7_to_datetime(UUID("2e92c8da-0b38-4d95-a627-d6e761f764f9"))
    assert str(ctx.value) == "UUIDv7 expected, received UUIDv4"



"""
Factories for new types.
"""

from collections import deque
from collections.abc import Callable
from typing import Any, Deque, Generic, TypeVar
from uuid import UUID

from .lastuuid import uuid7

__all__ = [
    "NewTypeFactory",
    "LastUUIDFactory",
]


T = TypeVar("T", bound=UUID)


class NewTypeFactory(Generic[T]):
    """
    Factory for NewType UUIDs
    """

    def __init__(self, newtype: Any, id_factory: Callable[[], UUID] = uuid7):
        """
        Create a factory of type that store the last instances.

        :param newtype: the type to be used
        :param id_factory: the factory to use, uuid7 by default.
        """
        self._newtype = newtype
        self._id_factory = id_factory

    def __call__(self, *args: Any, **kwargs: Any) -> T:
        """Generate a new value."""
        val = self._id_factory(*args, **kwargs)  # type: ignore
        if self._newtype:
            val: T = self._newtype(val)  # cast to NewType
        return val


class LastUUIDFactory(NewTypeFactory[T]):
    """
    Keep last UUID generated.

    ```python
    >>> from typing import NewType
    >>> from uuid import UUID
    >>> from lastuuid.utils import LastUUIDFactory
    >>> ClientId = NewType("ClientId", UUID)
    >>> client_id_factory = LastUUIDFactory[ClientId](ClientId)
    >>> client_id_factory()
    UUID('019b4f7d-f9d1-7d46-922a-7b83c4462366')
    >>> client_id_factory()
    UUID('019b4f7d-f9d2-7471-b6bb-c48f31146c56')
    >>> client_id_factory()
    UUID('019b4f7d-f9d2-7471-b6bb-c490f5b50e3a')
    >>> client_id_factory.last
    UUID('019b4f7d-f9d2-7471-b6bb-c490f5b50e3a')
    >>> client_id_factory.lasts[0]
    UUID('019b4f7d-f9d2-7471-b6bb-c490f5b50e3a')
    >>> client_id_factory.lasts[1]
    UUID('019b4f7d-f9d2-7471-b6bb-c48f31146c56')
    >>> client_id_factory.lasts[2]
    UUID('019b4f7d-f9d1-7d46-922a-7b83c4462366')
    ```
    """
    def __init__(
        self,
        newtype: Any,
        id_factory: Callable[[], UUID] = uuid7,
        cache_size: int = 10,
    ):
        """
        Create a factory of type that store the last instances.

        :param newtype: the type to be used
        :param cache_size: size of the queue that saved the last uuids
        """
        super().__init__(newtype, id_factory)
        self._cache: Deque[T] = deque(maxlen=cache_size)

    def __call__(self, *args: Any, **kwargs: Any) -> T:
        """Generate the id and cache it."""
        val = super().__call__(*args, **kwargs)
        self._cache.append(val)
        return val

    @property
    def last(self) -> T:
        """Most recently generated UUID-NewType instance."""
        return self._cache[-1]

    @property
    def lasts(self) -> list[T]:
        """
        Returns a list of the last N generated UUID-NewType instances.

        The list is ordered from most recent to oldest.
        The most recent value is accessible via `last`.
        """
        return list(reversed(self._cache))
